package jalview.schemes;

import jalview.datamodel.AnnotatedCollectionI;
import jalview.datamodel.HiddenMarkovModel;
import jalview.datamodel.SequenceCollectionI;
import jalview.datamodel.SequenceI;
import jalview.util.ColorUtils;
import jalview.util.Comparison;

import java.awt.Color;
import java.util.Map;

/**
 * Base class for colour schemes based on a selected Hidden Markov Model. The
 * colour is
 * <ul>
 * <li>white for a gap</li>
 * <li>red for an insertion</li>
 * <li>orange for negative information content</li>
 * <li>white to blue for increasing information content</li>
 * </ul>
 * where information content is the log ratio
 * 
 * <pre>
 *   log(profile match emission probability / residue background probability)
 * </pre>
 * 
 * Sub-class implementations use either global ('Uniprot') or local
 * ('alignment') background frequencies.
 * 
 * @author tzvanaalten
 * @author gmcarstairs
 */
public abstract class HmmerColourScheme extends ResidueColourScheme
{
  private static final Color INSERTION_COLOUR = new Color(230, 0, 0); // reddish

  private HiddenMarkovModel hmm;

  private Map<Character, Float> frequencies;

  /**
   * Constructor given a Hidden Markov Model
   * 
   * @param markov
   */
  public HmmerColourScheme(HiddenMarkovModel markov)
  {
    hmm = markov;
  }

  /**
   * Default constructor (required by ColourSchemes.loadColourSchemes)
   */
  public HmmerColourScheme()
  {
  }

  @Override
  public Color findColour(char symbol, int position, SequenceI seq,
          String consensusResidue, float pid)
  {
    return findColour(symbol, position);
  }

  /**
   * Returns the colour at a particular symbol at a column in the alignment:
   * <ul>
   * <li>white for a gap</li>
   * <li>red for an insertion</li>
   * <li>orange for negative information content</li>
   * <li>white to blue for increasing information content</li>
   * </ul>
   * 
   * @param symbol
   * @param column
   * @return
   */
  private Color findColour(char symbol, int column)
  {
    if (getHmm() == null || Comparison.isGap(symbol))
    {
      return Color.white;
    }
    if (Character.isLowerCase(symbol))
    {
      symbol = Character.toUpperCase(symbol);
    }

    double prob = getHmm().getMatchEmissionProbability(column, symbol);
    Float freq = 0f;

    if (!frequencies.containsKey(symbol))
    {
      return Color.WHITE;
    }
    else
    {
      freq = frequencies.get(symbol);
    }
    if (prob == 0D)
    {
      return INSERTION_COLOUR;
    }
    double value = Math.log(prob / freq.floatValue());
    Color colour = null;
    if (value > 0)
    {
      colour = ColorUtils.getGraduatedColour((float) value, 0,
              Color.WHITE, getMaxInformationScore(), Color.blue);
    }
    else if (value < 0)
    {
      return Color.ORANGE;
    }
    return colour;
  }

  /**
   * Answers the maximum possible value of information score (log ratio), for
   * use in scaling a graduated colour range
   * 
   * @return
   */
  abstract float getMaxInformationScore();

  @Override
  public void alignmentChanged(AnnotatedCollectionI collection,
          Map<SequenceI, SequenceCollectionI> hiddenReps)
  {
    /*
     * ? no need to do anything if alignment is adjusted
     * since findColour() handles everything 
     */
  }

  /**
   * Answers a new colour scheme instance based on the HMM of the first sequence
   * in ac that has an HMM
   */
  @Override
  public ColourSchemeI getInstance(AnnotatedCollectionI ac,
          Map<SequenceI, SequenceCollectionI> hiddenRepSequences)
  {
    SequenceI hmmSeq = ac.getHmmConsensus();
    HiddenMarkovModel model = hmmSeq == null ? null : hmmSeq.getHMM();

    return newInstance(ac, model);
  }

  /**
   * Answers a new instance of the colour scheme for the given HMM
   * 
   * @param ac
   * @param model
   * @return
   */
  protected abstract HmmerColourScheme newInstance(AnnotatedCollectionI ac,
          HiddenMarkovModel model);

  @Override
  public boolean isSimple()
  {
    return false;
  }

  /**
   * Answers true if the sequence collection has an HMM consensus sequence, else
   * false
   */
  @Override
  public boolean isApplicableTo(AnnotatedCollectionI ac)
  {
    return ac.getHmmConsensus() != null;
  }

  protected Map<Character, Float> getFrequencies()
  {
    return frequencies;
  }

  protected void setFrequencies(Map<Character, Float> frequencies)
  {
    this.frequencies = frequencies;
  }

  protected HiddenMarkovModel getHmm()
  {
    return hmm;
  }

}
